//
//  InputView.m
//  GHMMCurveRecognition
//
//  Created by Marcelo Cicconet on 10/20/11.
//  Copyright 2011 Georgia Tech. All rights reserved.
//

#import "InputView.h"

@implementation InputView

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        viewOrigin = NSMakePoint(frame.origin.x, frame.origin.y);
        viewWidth = frame.size.width;
        viewHeight = frame.size.height;
        maxPoints = 60000; // 10 min at 100 fps
        xPoints = (float *)malloc(maxPoints*sizeof(float));
        yPoints = (float *)malloc(maxPoints*sizeof(float));
        phase = 0;
        curveIndex = -1;
        maxCurves = 3;
        arrayOfCurves = [[NSMutableArray alloc] initWithCapacity:maxCurves];
//        arrayOfCurves2 = [[NSMutableArray alloc] initWithCapacity:maxCurves];
        
        rsRingBufferLength = 50;
        rsCurvePointsDistance = 0.05;
        rsRingBufferX = (float *)calloc(rsRingBufferLength, sizeof(float));
        rsRingBufferY = (float *)calloc(rsRingBufferLength, sizeof(float));
    }
    
    return self;
}

- (void)awakeFromNib
{
    [ringBufferView1 setUpWithLength:rsRingBufferLength];
    [ringBufferView2 setUpWithLength:rsRingBufferLength];
    [ringBufferView3 setUpWithLength:rsRingBufferLength];
    [maxCurveLengthTextField setIntValue:rsRingBufferLength];
}

- (void)dealloc
{
    free(rsRingBufferX);
    free(rsRingBufferY);
    [objGHMM release];
    [arrayOfCurves release];
//    [arrayOfCurves2 release];
    free(xPoints);
    free(yPoints);
    [super dealloc];
}

- (IBAction)setLeftIndex:(id)sender
{
    leftIndex = (int)((float)nPoints*[sender floatValue]/100.0);
    [curveLengthTextField setIntValue:rightIndex-leftIndex+1];
    [self setNeedsDisplay:YES];
}

- (IBAction)setRightIndex:(id)sender
{
    rightIndex = (int)((float)nPoints*[sender floatValue]/100.0);
    [curveLengthTextField setIntValue:rightIndex-leftIndex+1];
    [self setNeedsDisplay:YES];
}

- (IBAction)start:(id)sender
{
    phase = 1;
    nPoints = 0;
    curveIndex += 1;
    [curveIndexTextField setIntValue:curveIndex+1];
    [buttonStart setEnabled:NO];
    [buttonTrim setEnabled:YES];
    [self setNeedsDisplay:YES];
}

- (IBAction)trim:(id)sender
{
    phase = 2;
    leftIndex = 0;
    rightIndex = nPoints-1;
    [curveLengthTextField setIntValue:rightIndex-leftIndex+1];
    [buttonTrim setEnabled:NO];
    [buttonRecord setEnabled:YES];
    [self setNeedsDisplay:YES];
}

- (IBAction)record:(id)sender
{
    phase = 3;
    int numOfPoints = rightIndex-leftIndex+1;
    if (numOfPoints > rsRingBufferLength) {
        numOfPoints = rsRingBufferLength;
        rightIndex = leftIndex+rsRingBufferLength-1;
    }
    Curve * curve = [[Curve alloc] initWithNumOfPoints:numOfPoints xPoints:&xPoints[leftIndex] yPoints:&yPoints[leftIndex] reverseOrder:YES];
    [arrayOfCurves addObject:curve];
    
//    numOfPoints = numOfPoints-1;
//    float * xDirections = (float *)malloc(numOfPoints*sizeof(float));
//    float * yDirections = (float *)malloc(numOfPoints*sizeof(float));
//    for (int i = 0; i < numOfPoints; i++) {
//        float dx = curve.xPoints[i+1]-curve.xPoints[i];
//        float dy = curve.yPoints[i+1]-curve.yPoints[i];
//        float n = sqrtf(dx*dx+dy*dy);
//        dx = (dx/n+1.0)/2.0;
//        dy = (dy/n+1.0)/2.0;
//        xDirections[i] = dx;
//        yDirections[i] = dy;
//    }
//    Curve * curve2 = [[Curve alloc] initWithNumOfPoints:numOfPoints xPoints:xDirections yPoints:yDirections reverseOrder:YES];
//    [arrayOfCurves2 addObject:curve2];
    
//    [curve2 release];
//    free(xDirections);
//    free(yDirections);
    
    [curve release];
    
    [buttonRecord setEnabled:NO];
    if (curveIndex < 2) {
        [buttonStart setEnabled:YES];
    }
    if (curveIndex == 2) {
        [buttonTrain setEnabled:YES];
    }
    [self setNeedsDisplay:YES];
}

- (IBAction)train:(id)sender
{
    phase = 4;
    nPoints = 0;
    objGHMM = [[ObjGHMM alloc] init];
    [objGHMM trainWithCurves:arrayOfCurves];

//    objGHMM2 = [[ObjGHMM alloc] init];
//    [objGHMM2 trainWithCurves:arrayOfCurves2];
    
    switch ([arrayOfCurves count]) {
        case 1: {
            ringBufferView1.curveLength = [[arrayOfCurves objectAtIndex:0] numOfPoints];
            ringBufferView1.ID = 1;
            break;
        }
        case 2: {
            ringBufferView1.curveLength = [[arrayOfCurves objectAtIndex:0] numOfPoints];
            ringBufferView2.curveLength = [[arrayOfCurves objectAtIndex:1] numOfPoints];
            ringBufferView1.ID = 1;
            ringBufferView2.ID = 2;
            break;
        }
        case 3: {
            ringBufferView1.curveLength = [[arrayOfCurves objectAtIndex:0] numOfPoints];
            ringBufferView2.curveLength = [[arrayOfCurves objectAtIndex:1] numOfPoints];
            ringBufferView3.curveLength = [[arrayOfCurves objectAtIndex:2] numOfPoints];
            ringBufferView1.ID = 1;
            ringBufferView2.ID = 2;
            ringBufferView3.ID = 3;
            break;
        }
        default:
            break;
    }
    
    [buttonTrain setEnabled:NO];
    [buttonTest setEnabled:YES];
    [self setNeedsDisplay:YES];
}

- (IBAction)test:(id)sender
{
    phase = 5;
    nPoints = 0;
    nRsPoints = 0;
    [buttonTest setEnabled:NO];
    [self setNeedsDisplay:YES];
}

- (BOOL)acceptsFirstResponder {
    return YES;
}

- (void)compare
{
    Curve * curve = [[Curve alloc] initWithNumOfPoints:nRsPoints xPoints:rsRingBufferX yPoints:rsRingBufferY reverseOrder:NO];
    [objGHMM testWithCurve:curve];
    
//    int numOfPoints = nRsPoints-1;
//    float * xDirections = (float *)malloc(numOfPoints*sizeof(float));
//    float * yDirections = (float *)malloc(numOfPoints*sizeof(float));
//    for (int i = 0; i < numOfPoints; i++) {
//        float dx = curve.xPoints[i+1]-curve.xPoints[i];
//        float dy = curve.yPoints[i+1]-curve.yPoints[i];
//        float n = sqrtf(dx*dx+dy*dy);
//        dx = (dx/n+1.0)/2.0;
//        dy = (dy/n+1.0)/2.0;
//        xDirections[i] = dx;
//        yDirections[i] = dy;
//    }
//    Curve * curve2 = [[Curve alloc] initWithNumOfPoints:numOfPoints xPoints:xDirections yPoints:yDirections reverseOrder:NO];
//    [objGHMM2 testWithCurve:curve2];
//    [curve2 release];
//    free(xDirections);
//    free(yDirections);
    
    float d1 = 0.0;
    float d2 = 0.0;
    float d3 = 0.0;
    Curve * c1 = [arrayOfCurves objectAtIndex:0];
    Curve * c2 = [arrayOfCurves objectAtIndex:1];
    Curve * c3 = [arrayOfCurves objectAtIndex:2];
    float dx, dy;
    for (int i = 0; i < [c1 numOfPoints]; i++) {
        dx = c1.xPoints[i]-curve.xPoints[i];
        dy = c1.yPoints[i]-curve.yPoints[i];
        d1 += (dx*dx+dy*dy);
    }
    for (int i = 0; i < [c2 numOfPoints]; i++) {
        dx = c2.xPoints[i]-curve.xPoints[i];
        dy = c2.yPoints[i]-curve.yPoints[i];;
        d2 += (dx*dx+dy*dy);        
    }
    for (int i = 0; i < [c3 numOfPoints]; i++) {
        dx = c3.xPoints[i]-curve.xPoints[i];
        dy = c3.yPoints[i]-curve.yPoints[i];;
        d3 += (dx*dx+dy*dy);
    }
    d1 = 2*sqrtf(d1)/(float)[c1 numOfPoints];
    d2 = 2*sqrtf(d2)/(float)[c2 numOfPoints];
    d3 = 2*sqrtf(d3)/(float)[c3 numOfPoints];
    
    [curve release];
    
    switch ([arrayOfCurves count]) {
        case 1: {
            [ringBufferView1 newValue:objGHMM.distanceToCurves[0]
                            euclValue:d1
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[0]
                       directionValue:0.0
                              matches:objGHMM.maxStates[0]];
            break;
        }
        case 2: {
            [ringBufferView1 newValue:objGHMM.distanceToCurves[0]
                            euclValue:d1
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[0]
                       directionValue:0.0
                              matches:objGHMM.maxStates[0]];
            [ringBufferView2 newValue:objGHMM.distanceToCurves[1]
                            euclValue:d2
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[1]
                       directionValue:0.0
                              matches:objGHMM.maxStates[1]];
            break;
        }
        case 3: {
            [ringBufferView1 newValue:objGHMM.distanceToCurves[0]
                            euclValue:d1
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[0]
                       directionValue:0.0
                              matches:objGHMM.maxStates[0]];
            [ringBufferView2 newValue:objGHMM.distanceToCurves[1]
                            euclValue:d2
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[1]
                       directionValue:0.0
                              matches:objGHMM.maxStates[1]];
            [ringBufferView3 newValue:objGHMM.distanceToCurves[2]
                            euclValue:d3
                           alphaValue:objGHMM.distanceToCurvesUsingAlpha[2]
                       directionValue:0.0
                              matches:objGHMM.maxStates[2]];
            break;
        }
        default:
            break;
    }
}

- (void)resampleWithPoint:(NSPoint)thePoint
{
    if (nRsPoints == 0) {
        rsRingBufferX[0] = thePoint.x;
        rsRingBufferY[0] = thePoint.y;
        lastRsPoint = thePoint;
        nRsPoints += 1;
    } else {
        float d = [self distanceBetweenPointsP:thePoint andQ:lastRsPoint];
        if (d > rsCurvePointsDistance) {
            int nNewPoints = (int)(d/rsCurvePointsDistance);
            for (int point = 0; point < nNewPoints; point++) {
                for (int i = rsRingBufferLength-1; i > 0; i--) {
                    rsRingBufferX[i] = rsRingBufferX[i-1];
                    rsRingBufferY[i] = rsRingBufferY[i-1];
                }
                rsRingBufferX[0] = lastRsPoint.x+(point+1)*rsCurvePointsDistance*(thePoint.x-lastRsPoint.x)/d;
                rsRingBufferY[0] = lastRsPoint.y+(point+1)*rsCurvePointsDistance*(thePoint.y-lastRsPoint.y)/d;
                nRsPoints += 1;
                if (nRsPoints > rsRingBufferLength) {
                    nRsPoints = rsRingBufferLength;
                    [self compare];               
                }
            }
            lastRsPoint = NSMakePoint(rsRingBufferX[0], rsRingBufferY[0]);
        } // else: just ignore point
    }
}

- (float)distanceBetweenPointsP:(NSPoint)thePointP andQ:(NSPoint)thePointQ
{
    float dx = thePointP.x-thePointQ.x;
    float dy = thePointP.y-thePointQ.y;
    return sqrtf(dx*dx+dy*dy);
}

//- (void)mouseDown:(NSEvent *)theEvent {
//}

- (void)mouseDragged:(NSEvent *)theEvent
{
    if (phase == 1) {
        NSPoint p = [self locationInViewOfEvent:theEvent];
        p.x /= viewWidth;
        p.y /= viewHeight;
        if (nPoints == 0) {
            xPoints[nPoints] = p.x;
            yPoints[nPoints] = p.y;
            lastRsPoint = p;
            nPoints += 1;
        } else {
            float d = [self distanceBetweenPointsP:p andQ:lastRsPoint];
            if (d > rsCurvePointsDistance) {
                int nNewPoints = (int)(d/rsCurvePointsDistance);
                for (int point = 0; point < nNewPoints; point++) {
                    xPoints[nPoints] = lastRsPoint.x+(point+1)*rsCurvePointsDistance*(p.x-lastRsPoint.x)/d;
                    yPoints[nPoints] = lastRsPoint.y+(point+1)*rsCurvePointsDistance*(p.y-lastRsPoint.y)/d;
                    nPoints += 1;
                }
                lastRsPoint = NSMakePoint(xPoints[nPoints-1], yPoints[nPoints-1]);
            } // else: just ignore point
        }
        [self setNeedsDisplay:YES];
    } else if (phase == 5) {
        NSPoint p = [self locationInViewOfEvent:theEvent];
        p.x /= viewWidth;
        p.y /= viewHeight;
        [self resampleWithPoint:p];
        [self setNeedsDisplay:YES];
    }
}

//- (void)mouseUp:(NSEvent *)theEvent {
//}

- (NSPoint)locationInViewOfEvent:(NSEvent *)theEvent
{
	NSPoint p = [theEvent locationInWindow];
	return NSMakePoint(p.x-viewOrigin.x, p.y-viewOrigin.y);
}

- (void)drawRect:(NSRect)dirtyRect
{
    [[NSColor blackColor] set];
    [NSBezierPath fillRect:dirtyRect];
    switch (phase) {
        case 1: {
            [[NSColor yellowColor] set];
            for (int i = 0; i < nPoints; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            break;
        }
        case 2: {
            [[NSColor lightGrayColor] set];
            for (int i = 0; i < leftIndex; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            for (int i = rightIndex; i < nPoints; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            [[NSColor yellowColor] set];
            for (int i = leftIndex; i <= rightIndex; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            for (int i = leftIndex; i < rightIndex; i++) {
                [NSBezierPath strokeLineFromPoint:NSMakePoint(xPoints[i]*viewWidth, yPoints[i]*viewHeight)
                                          toPoint:NSMakePoint(xPoints[i+1]*viewWidth, yPoints[i+1]*viewHeight)];
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }            
            break;
        }
        case 3: {
            [[NSColor yellowColor] set];
            for (int i = leftIndex; i <= rightIndex; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            for (int i = leftIndex; i < rightIndex; i++) {
                [NSBezierPath strokeLineFromPoint:NSMakePoint(xPoints[i]*viewWidth, yPoints[i]*viewHeight)
                                          toPoint:NSMakePoint(xPoints[i+1]*viewWidth, yPoints[i+1]*viewHeight)];
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(xPoints[i]*viewWidth-2.0, yPoints[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }            
            break;
        }    
        case 5: {
//            [[NSColor yellowColor] set];
//            for (int i = 0; i < nPoints; i++) {
//                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(ringBufferX[i]*viewWidth-2.0, ringBufferY[i]*viewHeight-2.0, 4.0, 4.0)] fill];
//            }
            [[NSColor yellowColor] set];
            for (int i = 0; i < nRsPoints; i++) {
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(rsRingBufferX[i]*viewWidth-2.0, rsRingBufferY[i]*viewHeight-2.0, 4.0, 4.0)] fill];
            }
            break;
        }
        default:
            break;
    }
}

@end
