//
//  ModelGHMM.m
//  GHMMCurveRecognition
//
//  Created by Marcelo Cicconet on 10/21/11.
//  Copyright 2011 Georgia Tech. All rights reserved.
//

#import "ModelGHMM.h"

@implementation ModelGHMM

@synthesize n_states, states;

- (id)initWithCurve:(Curve *)theCurve
{
    self = [super init];
    if (self) {
        dimension = 2;
        n_states = [theCurve numOfPoints];
        
        c = calloc(1, sizeof(double));
        c[0] = 1.0;
        u = calloc(dimension*dimension, sizeof(double));
        inv_u = calloc(dimension*dimension, sizeof(double));
        chd_u = calloc(dimension*dimension, sizeof(double));
        u[0] = 0.5;
        u[1] = 0.0;
        u[2] = 0.0;
        u[3] = 0.5;
        double det = ighmm_determinant(u, 2);
        ighmm_inverse(u, 2, det, inv_u);
        ighmm_cholesky_decomposition(chd_u, 2, u);
        
        density = binormal;
        
        trans_prob_out_stencil_i = (double *)malloc(3*sizeof(double));
        trans_prob_out_stencil_Nm2 = (double *)malloc(2*sizeof(double));
        trans_prob_out_stencil_Nm1 = (double *)malloc(1*sizeof(double));
        trans_prob_in_stencil_0 = (double *)malloc(1*sizeof(double));
        trans_prob_in_stencil_1 = (double *)malloc(2*sizeof(double));
        trans_prob_in_stencil_i = (double *)malloc(3*sizeof(double));
        trans_prob_in_stencil_Nm1 = (double *)malloc(3*sizeof(double));
        
        trans_prob_out_stencil_i[0] = 1.0/3.0;
        trans_prob_out_stencil_i[1] = 1.0/3.0;
        trans_prob_out_stencil_i[2] = 1.0/3.0;
        
        trans_prob_out_stencil_Nm2[0] = 1.0/3.0;
        trans_prob_out_stencil_Nm2[1] = 2.0/3.0;
        
        trans_prob_out_stencil_Nm1[0] = 1.0;
        
        trans_prob_in_stencil_0[0] = 1.0/3.0;
        
        trans_prob_in_stencil_1[0] = 1.0/3.0;
        trans_prob_in_stencil_1[1] = 1.0/3.0;
        
        trans_prob_in_stencil_i[0] = 1.0/3.0;
        trans_prob_in_stencil_i[1] = 1.0/3.0;
        trans_prob_in_stencil_i[2] = 1.0/3.0;
        
        trans_prob_in_stencil_Nm1[0] = 1.0/3.0;
        trans_prob_in_stencil_Nm1[1] = 2.0/3.0;
        trans_prob_in_stencil_Nm1[2] = 1.0;
        
        
        emissions = (ghmm_c_emission *)malloc(n_states*sizeof(ghmm_c_emission));
        mu = (double **)malloc(n_states*sizeof(double *));
        for (int i = 0; i < n_states; i++) {
            mu[i] = (double *)malloc(2*sizeof(double));
            mu[i][0] = theCurve.xPoints[i];
            mu[i][1] = theCurve.yPoints[i];
//            printf("(%f, %f)\n", mu[i][0], mu[i][1]);
        }
//        printf("\n");
        for (int i = 0; i < n_states; i++) {
            emissions[i].type = density;
            ghmm_c_emission_alloc(&emissions[i], dimension);
            emissions[i].dimension = dimension;
            emissions[i].mean.vec[0] = mu[i][0];
            emissions[i].mean.vec[1] = mu[i][1];
            for (int j = 0; j < dimension*dimension; j++) {
                emissions[i].variance.mat[j] = u[j];
            }
            emissions[i].sigmainv = inv_u;
            emissions[i].sigmacd = chd_u;
            emissions[i].fixed = 1;
            emissions[i].det = det;
        }
        
        M = 1;
        trans_ids_out = (int **)malloc(n_states*sizeof(int *));
        trans_ids_in = (int **)malloc(n_states*sizeof(int *));
        for (int i = 0; i < n_states; i++) {
            if (i == 0) {
                trans_ids_out[i] = (int *)malloc(3*sizeof(int));
                trans_ids_in[i] = (int *)malloc(1*sizeof(int));
            } else if (i == 1) {
                trans_ids_out[i] = (int *)malloc(3*sizeof(int));
                trans_ids_in[i] = (int *)malloc(2*sizeof(int));
            } else if (i == n_states-2) {
                trans_ids_out[i] = (int *)malloc(2*sizeof(int));
                trans_ids_in[i] = (int *)malloc(3*sizeof(int));
            } else if (i == n_states-1) {
                trans_ids_out[i] = (int *)malloc(1*sizeof(int));
                trans_ids_in[i] = (int *)malloc(3*sizeof(int));
            } else {
                trans_ids_out[i] = (int *)malloc(3*sizeof(int));
                trans_ids_in[i] = (int *)malloc(3*sizeof(int));                
            }
        }
        for (int i = 0; i < n_states; i++) {
            if (i == 0) {
                trans_ids_out[i][0] = 0;
                trans_ids_out[i][1] = 1;
                trans_ids_out[i][2] = 2;
            } else if (i == 1) {
                trans_ids_out[i][0] = 1;
                trans_ids_out[i][1] = 2;
                trans_ids_out[i][2] = 3;
            } else if (i == n_states-2) {
                trans_ids_out[i][0] = n_states-2;
                trans_ids_out[i][1] = n_states-1;
            } else if (i == n_states-1) {
                trans_ids_out[i][0] = n_states-1;
            } else {
                trans_ids_out[i][0] = i;
                trans_ids_out[i][1] = i+1;
                trans_ids_out[i][2] = i+2;
            }
        }
        for (int j = 0; j < n_states; j++) {
            if (j == 0) {
                trans_ids_in[j][0] = 0;
            } else if (j == 1) {
                trans_ids_in[j][0] = 0;
                trans_ids_in[j][1] = 1;
            } else if (j == n_states-2) {
                trans_ids_in[j][0] = n_states-4;
                trans_ids_in[j][1] = n_states-3;
                trans_ids_in[j][2] = n_states-2;
            } else if (j == n_states-1) {
                trans_ids_in[j][0] = n_states-3;
                trans_ids_in[j][1] = n_states-2;
                trans_ids_in[j][2] = n_states-1;
            } else {
                trans_ids_in[j][0] = j-2;
                trans_ids_in[j][1] = j-1;
                trans_ids_in[j][2] = j;
            }
        }
        states = (ghmm_cstate *)malloc(n_states*sizeof(ghmm_cstate));
        for (int i = 0; i < n_states; i++) {
            states[i].M=M; // densities per state
            states[i].c=c; /* weight of distribution */
            states[i].e=&emissions[i]; // emission
            
//            if (i == 0) {
//                states[i].pi = 1.0/3.0;
//            } else if (i == 1) {
//                states[i].pi = 1.0/3.0;
//            } else if (i == 2) {
//                states[i].pi = 1.0/3.0;
//            } else {
//                states[i].pi = 0.0;
//            }
            
            states[i].pi = 1.0/(double)n_states; // initial probability
            
            states[i].out_id=trans_ids_out[i];
            states[i].in_id=trans_ids_in[i];
            
            if (i == 0) {
                states[i].out_states=3;
                states[i].in_states=1;
                states[i].out_a = &trans_prob_out_stencil_i;
                states[i].in_a = &trans_prob_in_stencil_0;
            } else if (i == 1) {
                states[i].out_states=3;
                states[i].in_states=2;
                states[i].out_a = &trans_prob_out_stencil_i;
                states[i].in_a = &trans_prob_in_stencil_1;
            } else if (i == n_states-2) {
                states[i].out_states=2;
                states[i].in_states=3;
                states[i].out_a = &trans_prob_out_stencil_Nm2;
                states[i].in_a = &trans_prob_in_stencil_i;
            } else if (i == n_states-1) {
                states[i].out_states=1;
                states[i].in_states=3;
                states[i].out_a = &trans_prob_out_stencil_Nm1;
                states[i].in_a = &trans_prob_in_stencil_Nm1;
            } else {
                states[i].out_states=3;
                states[i].in_states=3;
                states[i].out_a = &trans_prob_out_stencil_i;
                states[i].in_a = &trans_prob_in_stencil_i;
            }
            
            states[i].fix=1;
            states[i].desc = NULL;
        }
    }
    
    return self;
}

- (void)dealloc {
    free(trans_prob_out_stencil_i);
    free(trans_prob_out_stencil_Nm2);
    free(trans_prob_out_stencil_Nm1);
    free(trans_prob_in_stencil_0);
    free(trans_prob_in_stencil_1);
    free(trans_prob_in_stencil_i);
    free(trans_prob_in_stencil_Nm1);
    for (int i = 0; i < n_states; i++) {
        free(mu[i]);
    }
    free(mu);
    free(states);
    for (int i = 0; i < n_states; i++) {
        free(trans_ids_out[i]);
        free(trans_ids_in[i]);
    }
    free(trans_ids_out);
    free(trans_ids_in);
    for (int i = 0; i < n_states; i++) {
        ghmm_c_emission_free(&emissions[i]);
    }
    free(emissions);
    free(u);
    free(inv_u);
    free(chd_u);
    free(c);
    [super dealloc];
}

@end
