//
//  RingBufferView.m
//  GHMMCurveRecognition
//
//  Created by Marcelo Cicconet on 11/2/11.
//  Copyright (c) 2011 Georgia Tech. All rights reserved.
//

#import "RingBufferView.h"

@implementation RingBufferView

@synthesize ID, curveLength;

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code here.
    }
    
    return self;
}

- (void)setUpWithLength:(int)theLength
{
    // values should be in the range [0,1]
    length = theLength;
    values = (float *)calloc(length,sizeof(float));
    euclValues = (float *)calloc(length,sizeof(float));
    alphaValues = (float *)calloc(length,sizeof(float));
    directionValues = (float *)calloc(length,sizeof(float));
    threshold = 0.07;
}

- (void)newValue:(float)theValue
       euclValue:(float)theEuclValue
      alphaValue:(float)theAlphaValue
  directionValue:(float)theDirectionValue
         matches:(int *)theMatches
{
    for (int i = 0; i < length; i++) {
        values[i] = values[i+1];
        euclValues[i] = euclValues[i+1];
        alphaValues[i] = alphaValues[i+1];
        directionValues[i] = directionValues[i+1];
    }
    
    float smoothnessFactor = 0.3;
    
    values[length-1] = (1.0-smoothnessFactor)*theValue+smoothnessFactor*values[length-2];
    euclValues[length-1] = (1.0-smoothnessFactor)*theEuclValue+smoothnessFactor*euclValues[length-2];
    alphaValues[length-1] = (1.0-smoothnessFactor)*theAlphaValue+smoothnessFactor*alphaValues[length-2];
    directionValues[length-1] = (1.0-smoothnessFactor)*theDirectionValue+smoothnessFactor*directionValues[length-2];
    
    matches = theMatches;
    
    int nd;
    float accumulator = 0.0;
    for (int i = 0; i < curveLength-1; i++) {
        float diff = values[length-1-i]-values[length-2-i];
        if (diff >= 0) {
            nd = i;
            break;
        } else {
            accumulator -= diff;
        }
    }
//    if (ID == 1) {
//        printf("%d, %d\n", curveLength, nd);
//    }
    if (nd > 0.5*curveLength && accumulator > 0.2 && distanceToPreviousValley2 > 0.7*curveLength) {
        distanceToPreviousValley2 = 0;
    }
    distanceToPreviousValley2 += 1;
    
    
    if (/*values[length-1] < threshold &&*/ euclValues[length-1] < threshold && distanceToPreviousValley > 0.7*curveLength) {
        distanceToPreviousValley = 0;
    }
    distanceToPreviousValley += 1;
    
    [self setNeedsDisplay:YES];
}

- (void)dealloc
{
    free(values);
    free(euclValues);
    free(alphaValues);
    free(directionValues);
    [super dealloc];
}

- (void)drawRect:(NSRect)dirtyRect
{
    [[NSColor blackColor] set];
    [NSBezierPath fillRect:dirtyRect];
    [[NSColor redColor] set];
    [NSBezierPath strokeLineFromPoint:NSMakePoint(0.0, threshold*dirtyRect.size.height)
                              toPoint:NSMakePoint(dirtyRect.size.width, threshold*dirtyRect.size.height)];
    for (int i = 0; i < length-1; i++) {
        [[NSColor whiteColor] set];
        [NSBezierPath strokeLineFromPoint:NSMakePoint((float)i/(float)length*dirtyRect.size.width, values[i]*dirtyRect.size.height)
                                  toPoint:NSMakePoint((float)(i+1)/(float)length*dirtyRect.size.width, values[i+1]*dirtyRect.size.height)];
        [[NSColor blueColor] set];
        [NSBezierPath strokeLineFromPoint:NSMakePoint((float)i/(float)length*dirtyRect.size.width, euclValues[i]*dirtyRect.size.height)
                                  toPoint:NSMakePoint((float)(i+1)/(float)length*dirtyRect.size.width, euclValues[i+1]*dirtyRect.size.height)];
        [[NSColor yellowColor] set];
        [NSBezierPath strokeLineFromPoint:NSMakePoint((float)i/(float)length*dirtyRect.size.width, alphaValues[i]*dirtyRect.size.height)
                                  toPoint:NSMakePoint((float)(i+1)/(float)length*dirtyRect.size.width, alphaValues[i+1]*dirtyRect.size.height)];
        [[NSColor greenColor] set];
        [NSBezierPath strokeLineFromPoint:NSMakePoint((float)i/(float)length*dirtyRect.size.width, directionValues[i]*dirtyRect.size.height)
                                  toPoint:NSMakePoint((float)(i+1)/(float)length*dirtyRect.size.width, directionValues[i+1]*dirtyRect.size.height)];
    }
    [[NSColor blueColor] set];
    [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect((length-distanceToPreviousValley-1)/(float)length*dirtyRect.size.width-4.0, euclValues[length-distanceToPreviousValley-1]*dirtyRect.size.height-4.0, 8.0, 8.0)] fill];
    [[NSColor whiteColor] set];
    [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect((length-distanceToPreviousValley2-1)/(float)length*dirtyRect.size.width-4.0, values[length-distanceToPreviousValley2-1]*dirtyRect.size.height-4.0, 8.0, 8.0)] fill];
    
    if (matches) {
        [[NSColor grayColor] set];
        for (int i = 0; i < curveLength; i++) {
            float x = dirtyRect.size.width*(float)(i)/(float)(curveLength-1);
            float y = dirtyRect.size.height*(float)matches[i]/(float)(curveLength-1);
                [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(x-2.0, y-2.0, 4.0, 4.0)] stroke];
        }
    }
}

@end
